#include <stdlib.h>
#include <Windows.h>
#include "compilerfe.h"

using namespace win32fe;

compiler::compiler() {
  compileoutflag = linkoutflag = "-o ";

  OutputFlag = compilearg.end();

  OptionTags = "DILchlo";

  Options['D'] = &compiler::FoundD;
  Options['I'] = &compiler::FoundI;
  Options['L'] = &compiler::FoundL;
  Options['c'] = &compiler::Foundc;
  Options['h'] = &compiler::Foundh;
  Options['l'] = &compiler::Foundl;
  Options['o'] = &compiler::Foundo;
  Options[UNKNOWN] = &compiler::FoundUnknown;
}

int compiler::Parse(void) {
  if (!tool::Parse()) {
    LI i = arg.begin();
    compilearg.push_front(*i++);
    arg.pop_front();
    while (i != arg.end()) {
      std::string temp = *i;
      if (temp[0]!='-') {
        FoundFile(i);
      } else {
        char flag = temp[1];
        if (OptionTags.find(flag)==std::string::npos) {
          (this->*Options[UNKNOWN])(i);
        } else {
          (this->*Options[flag])(i);
        }
      }
      i++;
      arg.pop_front();
    }
  } else {
    return(-1);
  }
  return(0);
}

void compiler::AddSystemInfo(void) {
  if (detect) {
    tool::AddSystemInfo();
    AddSystemInclude();
    AddSystemLib();
  }
}

void compiler::AddSystemInclude(void) {
  /* System headers are in InstallDir/include */
  std::string include = InstallDir + "include";
  if (verbose) {
    std::string longpath;
    GetLongPath(include,longpath);
    std::cout << "win32fe: Adding Flag: -I" << longpath << std::endl;
  }
  arg.push_back("-I" + include);
  LI i = arg.end();
  FoundI(--i);
  arg.pop_back();
}

void compiler::AddSystemLib(void) {
  /* System libraries are in InstallDir/lib */ 
  std::string lib = InstallDir + "lib";
  if (verbose) {
    std::string longpath;
    GetLongPath(lib,longpath);
    std::cout << "win32fe: Adding Flag: -L" << longpath << std::endl;
  }
  arg.push_back("-L" + lib);
  LI i = arg.end();
  FoundL(--i);
  arg.pop_back();
}

int compiler::Execute(void) {
  int ierr=0;
  ierr = tool::Execute();
  if (!ierr && !(helpfound || versionfound)) {
    FixOutput();
    /* Determine if we are compiling, or linking */ 
    LI i=linkarg.begin();
    std::string temp = *i;
    if (temp == "-c") {
      ierr = Compile();
    } else {
      ierr = Link();
    }
  }
  return(ierr);
}

void compiler::Help(void) {
  tool::Help();
  std::cout << "For compilers:" << std::endl;
  std::cout << "  win32fe will map the following <tool options> to their native options:" << std::endl;
  std::cout << "    -c:          Compile Only, generates an object file with .o extension" << std::endl;
  std::cout << "                 This will invoke the compiler once for each file listed." << std::endl;
  std::cout << "    -l<library>: Link the file lib<library>.lib" << std::endl;
  std::cout << "    -o <file>:   Output=<file> context dependent" << std::endl;
  std::cout << "    -D<macro>:   Define <macro>" << std::endl;
  std::cout << "    -I<path>:    Add <path> to the include path" << std::endl;
  std::cout << "    -L<path>:    Add <path> to the link path" << std::endl;
  std::cout << "    -help:       <tool> specific help for win32fe" << std::endl << std::endl;
  std::cout << "Ex: win32fe cl -Zi -c foo.c --verbose -Iinclude" << std::endl << std::endl;
  std::cout << "Note: win32fe will automatically find the system library paths and" << std::endl;
  std::cout << "      system include paths, relieving the user of the need to invoke a" << std::endl;
  std::cout << "      particular shell.  The only requirement is that the compiler be" << std::endl;
  std::cout << "      found in the path, or have the path specified with --use <compiler>" << std::endl;
  std::cout << "      or --path <bin directory>." << std::endl << std::endl;
  std::cout << "=========================================================================" << std::endl << std::endl;
}

int compiler::Compile(void) {
  int n,ierr=0;
  LI i = compilearg.begin();
  std::string compile = *i++;
  Merge(compile,compilearg,i);

  /* Get the current working directory */
  std::string cwd;
  char directory[MAX_PATH]; /* Windows.h */
  GetCurrentDirectory(MAX_PATH,directory); /* Windows.h */
  cwd=(std::string)directory;
  GetShortPath(cwd);

  /* Execute each compilation one at a time */ 
  for (i=file.begin();i!=file.end();i++) {
    std::string outfile = *i;

    if (OutputFlag==compilearg.end()) {
      /* Make default output a .o not a .obj */
      n = outfile.rfind(".");
      outfile = outfile.substr(0,n) + ".o";
    } else {
      /* remove output file from compilearg list */
      outfile = *OutputFlag;
      compilearg.erase(OutputFlag);
      OutputFlag = compilearg.end();
      LI ii = compilearg.begin();
      compile = *ii++;
      Merge(compile,compilearg,ii);
    }

    /* Make sure output directory exists */ 
    std::string::size_type n = outfile.rfind("\\");
    if (n!=std::string::npos) {
      std::string dir = outfile.substr(0,n);
      if (GetShortPath(dir)) {
        outfile = dir + outfile.substr(n);
      } else {
        std::cerr << "Error: win32fe: Output Directory Not Found: ";
        std::cerr << outfile.substr(0,n) << std::endl;
        return(-1);
      }
    }
    outfile = compileoutflag + outfile;

    /* Concatenate the current directory with the file name if the file is local */
    std::string filename = *i;
    n = filename.find(":");
    if (n==std::string::npos) {
      n = filename.rfind("\\");
      if (n!=std::string::npos) {
        if (!GetShortPath(filename)) {
          std::cerr << "Error: win32fe: Input File Not Found: " << *i << std::endl;
          return(-1);
        }
      }
      filename = cwd + "\\" + filename;
    }
    std::string compileeach = compile + " " + outfile + " " + filename;
    if (verbose) {
      std::cout << compileeach << std::endl;
    }
    ierr = system(compileeach.c_str());
    if (ierr) break;
  }
  return(ierr);
}

int compiler::Link(void) {
  int ierr=0;
  LI i = compilearg.begin();
  std::string link = *i++;
  Merge(link,compilearg,i);
  i = file.begin();
  Merge(link,file,i);
  i = linkarg.begin();
  Merge(link,linkarg,i);
  if (verbose) {
    std::cout << link << std::endl;
  }
  ierr = system(link.c_str());
  return(ierr);
}

void compiler::FoundD(LI &i) {
  std::string temp = *i;
  ProtectQuotes(temp);
  compilearg.push_back(temp);
}

void compiler::FoundI(LI &i) {
  std::string shortpath = i->substr(2);
  ReplaceSlashWithBackslash(shortpath);
  if (GetShortPath(shortpath)) {
    shortpath = "-I"+shortpath;
    compilearg.push_back(shortpath);
  } else {
    if (!woff) {
      std::cerr << "Warning: win32fe Include Path Not Found: " << i->substr(2) << std::endl;
    }
  }
}

void compiler::FoundL(LI &i) {
  std::string shortpath =i->substr(2);
  ReplaceSlashWithBackslash(shortpath);
  if (GetShortPath(shortpath)) {
    shortpath = "-L"+shortpath;
    linkarg.push_back(shortpath);
  } else {
    if (!woff) {
      std::cerr << "Warning: win32fe Library Path Not Found: " << i->substr(2) << std::endl;
    }
  }
}

void compiler::Foundc(LI &i) {
  std::string temp = *i;
  if (temp=="-c") {
    compilearg.push_back(temp);
    linkarg.pop_front();
    linkarg.push_front(temp);
  } else {
    compilearg.push_back(temp);
  }
}

void compiler::Foundh(LI &i) {
  if (*i=="-help") {
    helpfound = true;
  }
}

void compiler::Foundl(LI &i) { 
  std::string temp = *i;
  file.push_back("lib" + temp.substr(2) + ".lib");
} 

void compiler::Foundo(LI &i) {
  if (*i == "-o") {
    i++;
    arg.pop_front();
    std::string temp = *i;
    ReplaceSlashWithBackslash(temp);
    compilearg.push_back(temp);
    /* Set Flag then fix later based on compilation or link */
    OutputFlag = --compilearg.end();
  } else {
    compilearg.push_back(*i);
  }
}   

void compiler::FoundUnknown(LI &i) {
  std::string temp = *i;
  compilearg.push_back(temp);
}

void compiler::FixOutput(void) {
  if (OutputFlag!=compilearg.end()) {
    std::string outfile = *OutputFlag;
    compilearg.erase(OutputFlag);
    if (linkarg.front()=="-c") {
      compilearg.push_back(outfile);
      OutputFlag = --compilearg.end();
    } else {
      outfile = linkoutflag + outfile;
      linkarg.push_front(outfile);
      OutputFlag = --linkarg.begin();
    }
  }
}

void compiler::DisplayVersion(void) {
  tool::DisplayVersion();
  std::string version = compilearg.front();
  version += " 2>&1 | head -1";
  system(version.c_str());
}

bool compiler::IsAKnownTool(void) {
  return(true);
}
